/*******************************************************************************
 * Copyright (c) 2009 Mia-Software.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Nicolas Payneau (Mia-Software) - initial API and implementation
 *******************************************************************************/
package org.eclipse.gmt.modisco.workflow.defaultengine;

import java.util.List;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.core.runtime.IExtension;
import org.eclipse.core.runtime.IExtensionPoint;
import org.eclipse.core.runtime.IExtensionRegistry;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.Platform;
import org.eclipse.debug.core.DebugPlugin;
import org.eclipse.debug.core.ILaunch;
import org.eclipse.debug.core.ILaunchConfiguration;
import org.eclipse.debug.core.ILaunchConfigurationWorkingCopy;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.gmt.modisco.infra.common.core.logging.MoDiscoLogger;
import org.eclipse.gmt.modisco.workflow.Activator;
import org.eclipse.gmt.modisco.workflow.Constants;
import org.eclipse.gmt.modisco.workflow.DriverExecutionInterface;
import org.eclipse.gmt.modisco.workflow.DriverInterface;
import org.eclipse.gmt.modisco.workflow.WorkflowExecutionInterface;
import org.eclipse.gmt.modisco.workflow.modiscoworkflow.Element;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.osgi.util.NLS;
import org.eclipse.swt.widgets.Shell;

/**
 * 
 * @author Nicolas Payneau
 */
public class WorkflowExecutionImpl implements WorkflowExecutionInterface {

	@SuppressWarnings("unchecked")
	public void launchWorkflow(final ILaunchConfiguration configuration,
			final String mode,
			final ILaunch launch, final IProgressMonitor monitor) {
		WorkflowModelImpl workflowImpl = new WorkflowModelImpl();
		workflowImpl.initialization(configuration, null);
		List<? extends EObject> listComponent = workflowImpl.getAllComponents();
		List<Element> listModiscoWork = (List<Element>) listComponent;
		for (Element modiscoWork : listModiscoWork) {
			DriverExecutionInterface driver = findPlugIn(modiscoWork.getType());
			if (driver != null) {
				ILaunchConfiguration conf = getLaunchConfigurationFromName(modiscoWork
						.getName());
				ILaunchConfigurationWorkingCopy config = null;
				try {
					config = conf.getWorkingCopy();
				} catch (CoreException e) {
					MoDiscoLogger.logError(e, Activator.getDefault());
				}
				driver.launch(config, launch);
			} else {
				MessageDialog.openError(new Shell(),
						Messages.WorkflowExecutionImpl_0, NLS.bind(
								Messages.WorkflowExecutionImpl_1,
								new Object[] { modiscoWork.getType() }));
			}
		}
	}

	/**
	 * Allows to retrieve a instance of {@link DriverExecutionInterface} for
	 * execute a configuration
	 * 
	 * @param type
	 *            The launcher type
	 * @return a instance of {@link DriverExecutionInterface} corresponding at
	 *         type specified
	 */
	private DriverExecutionInterface findPlugIn(final String type) {
		IExtensionRegistry registry = Platform.getExtensionRegistry();
		IExtensionPoint point = registry
				.getExtensionPoint(Constants.nameExtensionPointDriver);

		for (IExtension extension : point.getExtensions()) {
			for (IConfigurationElement element : extension
					.getConfigurationElements()) {
				if (element.getName().equals(Constants.nameElementDriver)) {
					try {
						DriverInterface driver = (DriverInterface) element
								.createExecutableExtension("class"); //$NON-NLS-1$
						if (driver.getTypeLaunchConfig().equals(type)) {
							return findExecutionPlugIn(extension
									.getUniqueIdentifier());
						}
					} catch (CoreException e) {
						MoDiscoLogger.logError(e, Activator.getDefault());
					}
				}
			}
		}
		return null;
	}

	/**
	 * Allows to retrieve a instance of {@link DriverExecutionInterface} for
	 * execute a configuration
	 * 
	 * @param idThe
	 *            id of plug-in specified
	 * @return a instance of {@link DriverExecutionInterface} corresponding at
	 *         id specified
	 */
	private DriverExecutionInterface findExecutionPlugIn(final String id) {
		IExtensionRegistry registry = Platform.getExtensionRegistry();
		IExtensionPoint point = registry
				.getExtensionPoint(Constants.nameExtensionPointDriver);

		for (IExtension extension : point.getExtensions()) {
			if (extension.getUniqueIdentifier().equals(id)) {
				for (IConfigurationElement element : extension
						.getConfigurationElements()) {
					if (element.getName()
							.equals(Constants.nameElementExecution)) {
						try {
							return (DriverExecutionInterface) element
									.createExecutableExtension("class"); //$NON-NLS-1$
						} catch (CoreException e) {
							MoDiscoLogger.logError(e, Activator.getDefault());
						}
					}
				}
			}
		}
		return null;
	}

	/**
	 * Allows to find the good instance of LaunchConfiguration corresponding at
	 * path
	 * 
	 * @param configName
	 *            Name of LaunchConfiguration
	 * @return An instance of launchConfiguration corresponding at path
	 */
	private ILaunchConfiguration getLaunchConfigurationFromName(
			final String configName) {
		ILaunchConfiguration configuration = null;
		try {
			for (ILaunchConfiguration config : DebugPlugin.getDefault()
					.getLaunchManager().getLaunchConfigurations()) {
				if (config.getName().equals(configName)) {
					configuration = config;
					break;
				}
			}
		} catch (CoreException e) {
			MoDiscoLogger.logError(e, Activator.getDefault());
		}
		return configuration;
	}

}
